use crate::{Object, Bytecode, Code, BOOL_TRUE, BOOL_FALSE, NULL, Frame};
use std::collections::HashMap;
use std::cell::RefCell;
use std::rc::Rc;

const STACK_SIZE: usize = 2048;

#[derive(Debug)]
pub struct VM {
    constants: Vec<Object>,

    frames: Vec<Frame>,
    frame: usize,

    globals: HashMap<usize, Object>,

    stack: Vec<Object>,
    sp: usize,

    last_popped: Option<Object>,
}

impl VM {

    fn current_frame(&mut self) -> &mut Frame {
        self.frames.get_mut(self.frame - 1).unwrap()
    }

    fn push_frame(&mut self, frame: Frame) {
        self.frames.push(frame);
        self.frame += 1;
    }

    fn pop_frame(&mut self) -> Frame {
        self.frame -= 1;

        self.frames.pop().unwrap()
    }

    fn top(&self) -> Option<&Object> {
        self.stack.get(self.sp).clone()
    }

    pub fn last_popped(&self) -> Option<Object> {
        self.last_popped.clone()
    }

    fn push(&mut self, object: Object) {
        if self.stack.len() > STACK_SIZE {
            panic!("Stack size exceeded.");
        }

        self.stack.push(object);
    }

    fn pop(&mut self) -> Object {
        self.stack.pop().unwrap()
    }

    pub fn run(&mut self) {
        while (self.current_frame().ip as usize) < self.current_frame().instructions.len() {
            let ip = self.current_frame().ip as usize;
            let op = self.current_frame().instructions.get(ip).unwrap().clone();

            match op {
                Code::SetGlobal(i) => {
                    self.current_frame().ip += 1;

                    let value = self.pop();
                    
                    self.globals.insert(i, value);
                },
                Code::SetLocal(i) => {
                    self.current_frame().ip += 1;

                    let base = self.current_frame().base;
                    let value = self.pop();

                    self.stack.insert(base + i, value);
                },
                Code::GetGlobal(i) => {
                    self.current_frame().ip += 1;

                    self.push(self.globals.get(&i).unwrap().clone())
                },
                Code::GetLocal(i) => {
                    self.current_frame().ip += 1;

                    let base = self.current_frame().base;
                    let value = self.stack.get(base + i).expect("Unable to get local variable.").clone();

                    self.push(value);
                },
                Code::True => {
                    self.push(BOOL_TRUE);
                    self.current_frame().ip += 1;
                },
                Code::False => {
                    self.push(BOOL_FALSE);
                    self.current_frame().ip += 1;
                },
                Code::Constant(index) => {
                    self.push(self.constants.get(index).expect("Unknown constant index.").clone());
                    self.current_frame().ip += 1;
                },
                Code::Add | Code::Subtract | Code::Mulitply | Code::Divide |
                Code::GreaterThan | Code::LessThan | Code::GreaterThanEquals | Code::LessThanEquals |
                Code::Equals | Code::NotEquals => {
                    let right = self.pop();
                    let left = self.pop();

                    let result = if matches!(op, Code::GreaterThan | Code::LessThan | Code::GreaterThanEquals | Code::LessThanEquals | Code::Equals | Code::NotEquals) {
                        match (left, right) {
                            (Object::Number(l), Object::Number(r)) => numeric_boolean_operation(l, op, r),
                            _ => todo!()
                        }
                    } else {
                        match (left, right) {
                            (Object::Number(l), Object::Number(r)) => numeric_binary_operation(l, op, r),
                            (Object::String(l), Object::String(r)) => string_binary_operation(l, op, r),
                            _ => todo!(),
                        }
                    };

                    self.push(result);

                    self.current_frame().ip += 1;
                },
                Code::Jump(pos) => {
                    self.current_frame().ip = pos;
                },
                Code::JumpIfNot(pos) => {
                    let object = self.pop();

                    if !object.to_bool() {
                        self.current_frame().ip = pos;
                    } else {
                        self.current_frame().ip += 1;
                    }
                },
                Code::Pop => {
                    let object = self.pop();
                    self.last_popped = Some(object);
                    self.current_frame().ip += 1;
                },
                Code::Negate => {
                    let object = self.pop();

                    self.push(Object::Number(match object {
                        Object::Number(n) => -n,
                        _ => todo!()
                    }));

                    self.current_frame().ip += 1;
                },
                Code::Return(with_value) => {
                    let value = if with_value { self.pop() } else { NULL };

                    let frame = self.pop_frame();

                    // This will tell us how many things were pushed to the stack during execution.
                    let mut number_of_locals = self.stack.len() - frame.base;

                    // Then we can go through and clean up the stack after execution.
                    while number_of_locals > 0 {
                        self.pop();
                        number_of_locals -= 1;
                    }

                    self.pop();
                    self.push(value);

                    self.current_frame().ip += 1;
                },
                Code::Call(number_of_args) => {
                    let (function, number_of_locals, number_of_params) = match self.stack.get(self.stack.len() - 1 - number_of_args).unwrap() {
                        Object::Function(instructions, number_of_locals, number_of_params) => (instructions.clone(), number_of_locals.clone(), number_of_params.clone()),
                        _ => unreachable!("calling non-function")
                    };

                    if number_of_args != number_of_params {
                        panic!("wrong number of arguments: wanted {}, got {}", number_of_params, number_of_args);
                    }

                    let frame = Frame::new(function, self.stack.len() - number_of_args);

                    self.push_frame(frame);
                },
                Code::Array(count) => {
                    let items = Rc::new(RefCell::new(Vec::new()));
                    
                    for i in 0..count {
                        items.borrow_mut().push(self.pop());
                    }

                    items.borrow_mut().reverse();
                    
                    self.push(Object::Array(items));
                    self.current_frame().ip += 1;
                },
                Code::GetProperty => {
                    let target = self.pop();

                    if ! matches!(target, Object::Array(_)) {
                        panic!("property / index access is only allowed on arrays.");
                    }

                    let property = self.pop();

                    self.current_frame().ip += 1;

                    match target {
                        Object::Array(items) => {
                            let property = property.to_usize();

                            let value = items.borrow().get(property).cloned().expect(
                                &format!("Undefined index: {}", property)
                            );

                            self.push(value);
                        },
                        _ => todo!("property access on {:?}", target),
                    }
                },
                Code::SetProperty => {
                    let value = self.pop();
                    let property = self.pop();
                    let target = self.pop();

                    if ! matches!(target, Object::Array(_)) {
                        panic!("property / index access is only allowed on arrays.");
                    }

                    self.current_frame().ip += 1;

                    match target {
                        Object::Array(items) => {
                            let property = property.to_usize();

                            items.borrow_mut()[property] = value.clone();

                            self.push(value);
                        },
                        _ => todo!("property access on {:?}", target),
                    }
                },
                _ => todo!("{:?}", op)
            };
        }
    }
}

#[allow(dead_code)]
fn numeric_binary_operation(left: f64, code: Code, right: f64) -> Object {
    Object::Number(match code {
        Code::Add => left + right,
        Code::Subtract => left - right,
        Code::Mulitply => left * right,
        Code::Divide => left / right,
        _ => todo!()
    })
}

#[allow(dead_code)]
fn numeric_boolean_operation(left: f64, code: Code, right: f64) -> Object {
    Object::Bool(match code {
        Code::LessThan => left < right,
        Code::GreaterThan => left > right,
        Code::LessThanEquals => left <= right,
        Code::GreaterThanEquals => left >= right,
        Code::Equals => left == right,
        Code::NotEquals => left != right,
        _ => todo!()
    })
}

#[allow(dead_code)]
fn string_binary_operation(left: String, code: Code, right: String) -> Object {
    Object::String(match code {
        Code::Add => left + &right,
        _ => panic!("incompatible operand types: string {:?} string", code),
    })
}

impl From<Bytecode> for VM {
    fn from(b: Bytecode) -> Self {
        let frame = Frame::new(b.instructions, 0);
        let frames = vec![frame];

        Self {
            frames: frames,
            frame: 1,

            constants: b.constants,

            stack: Vec::with_capacity(STACK_SIZE),
            sp: 0,

            globals: HashMap::new(),

            last_popped: None,
        }
    }
}

#[cfg(test)]
mod tests {
    use super::*;
    use std::cell::RefCell;
    use crate::utils::compile;

    #[test]
    fn arrays() {
        assert_eq!(parse_compile_and_run("[1];"), Object::Array(
            Rc::new(
                RefCell::new(vec![
                    Object::Number(1.0)
                ])
            )
        ).some());

        assert_eq!(parse_compile_and_run("[1][0];"), Object::Number(1.0).some());

        assert_eq!(parse_compile_and_run("
            let a = [1];
            a[0] = 2;
            a[0];
        "), Object::Number(2.0).some());
    }

    #[test]
    fn assigns() {
        assert_eq!(parse_compile_and_run("
            let foo = 2;
            foo = 3;
        "), Object::Number(3.0).some());
    }

    #[test]
    fn infixes() {
        assert_eq!(parse_compile_and_run("1 + 1;"), Object::Number(2.0).some());
        assert_eq!(parse_compile_and_run("1 - 1;"), Object::Number(0.0).some());
        assert_eq!(parse_compile_and_run("1 * 1;"), Object::Number(1.0).some());
        assert_eq!(parse_compile_and_run("1 / 1;"), Object::Number(1.0).some());

        assert_eq!(parse_compile_and_run("1 > 2;"), BOOL_FALSE.some());
        assert_eq!(parse_compile_and_run("1 < 2;"), BOOL_TRUE.some());
        assert_eq!(parse_compile_and_run("1 >= 1;"), BOOL_TRUE.some());
        assert_eq!(parse_compile_and_run("1 <= 1;"), BOOL_TRUE.some());
        assert_eq!(parse_compile_and_run("1 + 2 == 6 / 2;"), BOOL_TRUE.some());
        assert_eq!(parse_compile_and_run("1 + 2 != 9 / 2;"), BOOL_TRUE.some());
    }

    #[test]
    fn prefixes() {
        assert_eq!(parse_compile_and_run("-1;"), Object::Number(-1.0).some());
    }

    #[test]
    fn bools() {
        assert_eq!(parse_compile_and_run("true;"), Object::Bool(true).some());
        assert_eq!(parse_compile_and_run("false;"), Object::Bool(false).some());
    }

    #[test]
    fn conditionals() {
        assert_eq!(parse_compile_and_run("if true { 10; }"), Object::Number(10.0).some());
        assert_eq!(parse_compile_and_run("if false { 10; }"), None);
        assert_eq!(parse_compile_and_run("if true { 10; } else { 20; }"), Object::Number(10.0).some());
        assert_eq!(parse_compile_and_run("if false { 10; } else { 20; }"), Object::Number(20.0).some());
    }

    #[test]
    fn globals() {
        assert_eq!(parse_compile_and_run("let foo = 1; foo;"), Object::Number(1.0).some());
    }

    #[test]
    fn strings() {
        assert_eq!(parse_compile_and_run(r##""foo";"##), Object::String("foo".into()).some());
        assert_eq!(parse_compile_and_run(r##""foo" + "bar";"##), Object::String("foobar".into()).some());
        assert_eq!(parse_compile_and_run(r##""foo" + "bar" + "baz";"##), Object::String("foobarbaz".into()).some());
    }
    
    #[test]
    fn functions() {
        assert_eq!(parse_compile_and_run("fn foo() { return 5 + 10; } foo();"), Object::Number(15.0).some());
        assert_eq!(parse_compile_and_run("fn foo() { return 5; } foo() + foo();"), Object::Number(10.0).some());
        assert_eq!(parse_compile_and_run("fn foo() { return 5 + 10; } foo() + foo();"), Object::Number(30.0).some());
        assert_eq!(parse_compile_and_run("fn foo() {} foo();"), Object::Null.some());

        assert_eq!(parse_compile_and_run("
            fn a() { return 1; }
            fn b() { return a() + 1; }
            fn c() { return b() + 2; }
            c();
        "), Object::Number(4.0).some());

        assert_eq!(parse_compile_and_run("
            fn a() { return 1; }
            fn b() { return a; }
            b()();
        "), Object::Number(1.0).some());

        assert_eq!(parse_compile_and_run("
            fn a() { let foo = 1; return foo; }
            a();
        "), Object::Number(1.0).some());

        assert_eq!(parse_compile_and_run("
            fn a() { let foo = 1; let bar = 2; return foo + bar; }
            a();
        "), Object::Number(3.0).some());

        assert_eq!(parse_compile_and_run("
            let foo = 1;

            fn a() {
                let num = 1;
                return foo - num;
            }

            fn b() {
                let num = 2;
                return foo * num;
            }

            a() + b();
        "), Object::Number(2.0).some());

        assert_eq!(parse_compile_and_run("
            fn b() {
                return 1;
            }

            fn a() {
                return b;    
            }

            a()();
        "), Object::Number(1.0).some());

        assert_eq!(parse_compile_and_run("
            fn foo(bar) {
                return bar;
            }

            foo(1);
        "), Object::Number(1.0).some());

        assert_eq!(parse_compile_and_run("
            fn add(a, b) {
                return a + b;
            }

            add(1, 2);
        "), Object::Number(3.0).some());

        assert_eq!(parse_compile_and_run("
            fn with(value, callback) {
                return callback(value);
            }

            fn increment(value) {
                return 1 + value;
            }

            with(5, increment);
        "), Object::Number(6.0).some());

        assert_eq!(parse_compile_and_run("
            let global = 10;

            fn sum(a, b) {
                let c = a + b;
                return c;
            }

            fn outer() {
                return sum(1, 2) + sum(3, 4);
            }

            outer() + global;
        "), Object::Number(20.0).some());
    }

    fn parse_compile_and_run(source: &str) -> Option<Object> {
        let compiler = compile(source);
        let bytecode: Bytecode = compiler.into();

        let mut vm = VM::from(bytecode);

        vm.run();
        vm.last_popped()
    }
}